import re
import pytest

from ssh_audit.readbuf import ReadBuf
from ssh_audit.writebuf import WriteBuf


# pylint: disable=attribute-defined-outside-init,bad-whitespace
class TestBuffer:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.rbuf = ReadBuf
        self.wbuf = WriteBuf
        self.utf8rchar = b'\xef\xbf\xbd'

    @classmethod
    def _b(cls, v):
        v = re.sub(r'\s', '', v)
        data = [int(v[i * 2:i * 2 + 2], 16) for i in range(len(v) // 2)]
        return bytes(bytearray(data))

    def test_unread(self):
        w = self.wbuf().write_byte(1).write_int(2).write_flush()
        r = self.rbuf(w)
        assert r.unread_len == 5
        r.read_byte()
        assert r.unread_len == 4
        r.read_int()
        assert r.unread_len == 0

    def test_byte(self):
        w = lambda x: self.wbuf().write_byte(x).write_flush()  # noqa
        r = lambda x: self.rbuf(x).read_byte()  # noqa
        tc = [(0x00, '00'),
              (0x01, '01'),
              (0x10, '10'),
              (0xff, 'ff')]
        for p in tc:
            assert w(p[0]) == self._b(p[1])
            assert r(self._b(p[1])) == p[0]

    def test_bool(self):
        w = lambda x: self.wbuf().write_bool(x).write_flush()  # noqa
        r = lambda x: self.rbuf(x).read_bool()  # noqa
        tc = [(True,  '01'),
              (False, '00')]
        for p in tc:
            assert w(p[0]) == self._b(p[1])
            assert r(self._b(p[1])) == p[0]

    def test_int(self):
        w = lambda x: self.wbuf().write_int(x).write_flush()  # noqa
        r = lambda x: self.rbuf(x).read_int()  # noqa
        tc = [(0x00,       '00 00 00 00'),
              (0x01,       '00 00 00 01'),
              (0xabcd,     '00 00 ab cd'),
              (0xffffffff, 'ff ff ff ff')]
        for p in tc:
            assert w(p[0]) == self._b(p[1])
            assert r(self._b(p[1])) == p[0]

    def test_string(self):
        w = lambda x: self.wbuf().write_string(x).write_flush()  # noqa
        r = lambda x: self.rbuf(x).read_string()  # noqa
        tc = [('abc1',  '00 00 00 04 61 62 63 31'),
              (b'abc2',  '00 00 00 04 61 62 63 32')]
        for p in tc:
            v = p[0]
            assert w(v) == self._b(p[1])
            if not isinstance(v, bytes):
                v = bytes(bytearray(v, 'utf-8'))
            assert r(self._b(p[1])) == v

    def test_list(self):
        w = lambda x: self.wbuf().write_list(x).write_flush()  # noqa
        r = lambda x: self.rbuf(x).read_list()  # noqa
        tc = [(['d', 'ef', 'ault'], '00 00 00 09 64 2c 65 66 2c 61 75 6c 74')]
        for p in tc:
            assert w(p[0]) == self._b(p[1])
            assert r(self._b(p[1])) == p[0]

    def test_list_nonutf8(self):
        r = lambda x: self.rbuf(x).read_list()  # noqa
        src = self._b('00 00 00 04 de ad be ef')
        dst = [(b'\xde\xad' + self.utf8rchar + self.utf8rchar).decode('utf-8')]
        assert r(src) == dst

    def test_line(self):
        w = lambda x: self.wbuf().write_line(x).write_flush()  # noqa
        r = lambda x: self.rbuf(x).read_line()  # noqa
        tc = [('example line', '65 78 61 6d 70 6c 65 20 6c 69 6e 65 0d 0a')]
        for p in tc:
            assert w(p[0]) == self._b(p[1])
            assert r(self._b(p[1])) == p[0]

    def test_line_nonutf8(self):
        r = lambda x: self.rbuf(x).read_line()  # noqa
        src = self._b('de ad be af')
        dst = (b'\xde\xad' + self.utf8rchar + self.utf8rchar).decode('utf-8')
        assert r(src) == dst

    def test_bitlen(self):
        # pylint: disable=protected-access
        class Py26Int(int):
            def bit_length(self):
                raise AttributeError
        assert self.wbuf._bitlength(42) == 6
        assert self.wbuf._bitlength(Py26Int(42)) == 6

    def test_mpint1(self):
        mpint1w = lambda x: self.wbuf().write_mpint1(x).write_flush()  # noqa
        mpint1r = lambda x: self.rbuf(x).read_mpint1()  # noqa
        tc = [(0x0,     '00 00'),
              (0x1234,  '00 0d 12 34'),
              (0x12345, '00 11 01 23 45'),
              (0xdeadbeef, '00 20 de ad be ef')]
        for p in tc:
            assert mpint1w(p[0]) == self._b(p[1])
            assert mpint1r(self._b(p[1])) == p[0]

    def test_mpint2(self):
        mpint2w = lambda x: self.wbuf().write_mpint2(x).write_flush()  # noqa
        mpint2r = lambda x: self.rbuf(x).read_mpint2()  # noqa
        tc = [(0x0,               '00 00 00 00'),
              (0x80,              '00 00 00 02 00 80'),
              (0x9a378f9b2e332a7, '00 00 00 08 09 a3 78 f9 b2 e3 32 a7'),
              (-0x1234,           '00 00 00 02 ed cc'),
              (-0xdeadbeef,       '00 00 00 05 ff 21 52 41 11'),
              (-0x8000,           '00 00 00 02 80 00'),
              (-0x80,             '00 00 00 01 80')]
        for p in tc:
            assert mpint2w(p[0]) == self._b(p[1])
            assert mpint2r(self._b(p[1])) == p[0]
        assert mpint2r(self._b('00 00 00 02 ff 80')) == -0x80

    def test_reset(self):
        w = self.wbuf()
        w.write_int(7)
        w.write_int(13)
        assert len(w.write_flush()) == 8

        w.write_int(7)
        w.write_int(13)
        w.reset()
        assert len(w.write_flush()) == 0



import pytest


# pylint: disable=attribute-defined-outside-init
class TestUtils:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.utils = ssh_audit.Utils

    def test_to_bytes(self):
        assert self.utils.to_bytes(b'fran\xc3\xa7ais') == b'fran\xc3\xa7ais'
        assert self.utils.to_bytes('fran\xe7ais') == b'fran\xc3\xa7ais'
        # other
        with pytest.raises(TypeError):
            self.utils.to_bytes(123)

    def test_to_text(self):
        assert self.utils.to_text(b'fran\xc3\xa7ais') == 'fran\xe7ais'
        assert self.utils.to_text('fran\xe7ais') == 'fran\xe7ais'
        # other
        with pytest.raises(TypeError):
            self.utils.to_text(123)

    def test_is_ascii(self):
        assert self.utils.is_ascii('francais') is True
        assert self.utils.is_ascii('fran\xe7ais') is False
        # other
        assert self.utils.is_ascii(123) is False

    def test_to_ascii(self):
        assert self.utils.to_ascii('francais') == 'francais'
        assert self.utils.to_ascii('fran\xe7ais') == 'fran?ais'
        assert self.utils.to_ascii('fran\xe7ais', 'ignore') == 'franais'
        with pytest.raises(TypeError):
            self.utils.to_ascii(123)

    def test_is_print_ascii(self):
        assert self.utils.is_print_ascii('francais') is True
        assert self.utils.is_print_ascii('francais\n') is False
        assert self.utils.is_print_ascii('fran\xe7ais') is False
        # other
        assert self.utils.is_print_ascii(123) is False

    def test_to_print_ascii(self):
        assert self.utils.to_print_ascii('francais') == 'francais'
        assert self.utils.to_print_ascii('francais\n') == 'francais?'
        assert self.utils.to_print_ascii('fran\xe7ais') == 'fran?ais'
        assert self.utils.to_print_ascii('fran\xe7ais\n') == 'fran?ais?'
        assert self.utils.to_print_ascii('fran\xe7ais', 'ignore') == 'franais'
        assert self.utils.to_print_ascii('fran\xe7ais\n', 'ignore') == 'franais'
        with pytest.raises(TypeError):
            self.utils.to_print_ascii(123)

    def test_ctoi(self):
        assert self.utils.ctoi(123) == 123
        assert self.utils.ctoi('ABC') == 65

    def test_parse_int(self):
        assert self.utils.parse_int(123) == 123
        assert self.utils.parse_int('123') == 123
        assert self.utils.parse_int(-123) == -123
        assert self.utils.parse_int('-123') == -123
        assert self.utils.parse_int('abc') == 0

    def test_unique_seq(self):
        assert self.utils.unique_seq((1, 2, 2, 3, 3, 3)) == (1, 2, 3)
        assert self.utils.unique_seq((3, 3, 3, 2, 2, 1)) == (3, 2, 1)
        assert self.utils.unique_seq([1, 2, 2, 3, 3, 3]) == [1, 2, 3]
        assert self.utils.unique_seq([3, 3, 3, 2, 2, 1]) == [3, 2, 1]

    def test_parse_float(self):
        assert self.utils.parse_float('5.x') == -1.0

    def test_ipv6address(self):
        assert self.utils.is_ipv6_address('1.2.3.4') is False
        assert self.utils.is_ipv6_address('2600:1f18:420a:b500:bc4:c9c6:1d6:e3e4') is True



import hashlib
import pytest
from datetime import date

from ssh_audit.outputbuffer import OutputBuffer
from ssh_audit.policy import Policy
from ssh_audit.ssh2_kex import SSH2_Kex
from ssh_audit.writebuf import WriteBuf


class TestPolicy:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.OutputBuffer = OutputBuffer
        self.Policy = Policy
        self.wbuf = WriteBuf
        self.ssh2_kex = SSH2_Kex


    def _get_kex(self):
        '''Returns an SSH2.Kex object to simulate a server connection.'''

        w = self.wbuf()
        w.write(b'\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff')
        w.write_list(['kex_alg1', 'kex_alg2'])
        w.write_list(['key_alg1', 'key_alg2'])
        w.write_list(['cipher_alg1', 'cipher_alg2', 'cipher_alg3'])
        w.write_list(['cipher_alg1', 'cipher_alg2', 'cipher_alg3'])
        w.write_list(['mac_alg1', 'mac_alg2', 'mac_alg3'])
        w.write_list(['mac_alg1', 'mac_alg2', 'mac_alg3'])
        w.write_list(['comp_alg1', 'comp_alg2'])
        w.write_list(['comp_alg1', 'comp_alg2'])
        w.write_list([''])
        w.write_list([''])
        w.write_byte(False)
        w.write_int(0)
        return self.ssh2_kex.parse(self.OutputBuffer, w.write_flush())


    def test_builtin_policy_consistency(self):
        '''Ensure that the BUILTIN_POLICIES struct is consistent.'''

        for policy_name in Policy.BUILTIN_POLICIES:
            # Ensure that the policy name ends with " (version X)", where X is the 'version' field.
            version_str = " (version %s)" % Policy.BUILTIN_POLICIES[policy_name]['version']
            assert policy_name.endswith(version_str)

            # Ensure that all required fields are present.
            required_fields = ['version', 'banner', 'compressions', 'host_keys', 'optional_host_keys', 'kex', 'ciphers', 'macs', 'hostkey_sizes', 'dh_modulus_sizes', 'server_policy']
            for field in required_fields:
                assert field in Policy.BUILTIN_POLICIES[policy_name]

            # Ensure no extra fields are present.
            assert len(required_fields) == len(Policy.BUILTIN_POLICIES[policy_name])

            # Ensure that at least one host key is defined.
            assert type(Policy.BUILTIN_POLICIES[policy_name]['host_keys']) is list
            assert len(Policy.BUILTIN_POLICIES[policy_name]['host_keys']) > 0

            # Ensure that at least one key exchange is defined.
            assert type(Policy.BUILTIN_POLICIES[policy_name]['kex']) is list
            assert len(Policy.BUILTIN_POLICIES[policy_name]['kex']) > 0

            # Ensure that at least one cipher is defined.
            assert type(Policy.BUILTIN_POLICIES[policy_name]['ciphers']) is list
            assert len(Policy.BUILTIN_POLICIES[policy_name]['ciphers']) > 0

            # Ensure that at least one MAC is defined
            assert type(Policy.BUILTIN_POLICIES[policy_name]['macs']) is list
            assert len(Policy.BUILTIN_POLICIES[policy_name]['macs']) > 0

            # These tests apply to server policies only.
            if Policy.BUILTIN_POLICIES[policy_name]['server_policy']:
                assert type(Policy.BUILTIN_POLICIES[policy_name]['hostkey_sizes']) is dict
                assert len(Policy.BUILTIN_POLICIES[policy_name]['hostkey_sizes']) > 0

                # Examine all the hostkey_sizes entries...
                for hostkey_type in Policy.BUILTIN_POLICIES[policy_name]['hostkey_sizes']:
                    hostkey_data = Policy.BUILTIN_POLICIES[policy_name]['hostkey_sizes'][hostkey_type]

                    # Ensure that 'hostkey_size' is always included and that it is an integer.
                    assert 'hostkey_size' in hostkey_data
                    assert type(hostkey_data['hostkey_size']) is int

                    # If this is an ed25519 host key, ensure its size is fixed at 256.  If its an RSA host key, ensure the size is 4096.
                    if hostkey_type.find('ed25519') != -1:
                        assert int(hostkey_data['hostkey_size']) == 256
                    elif hostkey_type.startswith('rsa-'):
                        assert int(hostkey_data['hostkey_size']) == 4096
                    else:  # Catch unexpected host key types.
                        assert False

                    # Ensure either that 'ca_key_type' and 'ca_key_size' are both present, or neither are.  Fail cases when only one of the fields are present.
                    assert (('ca_key_type' in hostkey_data) and ('ca_key_size' in hostkey_data)) or (('ca_key_type' not in hostkey_data) and ('ca_key_size' not in hostkey_data))

                    # Ensure that the ca_key_type is either ssh-rsa or ssh-ed25519.
                    if 'ca_key_type' in hostkey_data:
                        assert hostkey_data['ca_key_type'] in ['ssh-rsa', 'ssh-ed25519']

                    # Ensure RSA CA key sizes are fixed at 4096.  Ensure ED25519 CA keys are 256.
                    if 'ca_key_size' in hostkey_data:
                        if 'ca_key_type' == 'ssh-rsa':
                            assert hostkey_data['ca_key_size'] == 4096
                        elif 'ca_key_type' == 'ssh-ed25519':
                            assert hostkey_data['ca_key_size'] == 256

                # Ensure that the 'dh_modulus_size' field is a dict.
                assert type(Policy.BUILTIN_POLICIES[policy_name]['dh_modulus_sizes']) is dict

                # The 'dh_modulus_size' field should have either one entry, or be empty.
                assert len(Policy.BUILTIN_POLICIES[policy_name]['dh_modulus_sizes']) in range(0, 2)  # The endpoint in range() is not inclusive

                # If 'diffie-hellman-group-exchange-sha256' is in the kex list, ensure that it exists in the 'dh_modulus_sizes' entry.  That entry must be defined for 2048 bits or larger.
                if 'diffie-hellman-group-exchange-sha256' in Policy.BUILTIN_POLICIES[policy_name]['kex']:
                    assert 'diffie-hellman-group-exchange-sha256' in Policy.BUILTIN_POLICIES[policy_name]['dh_modulus_sizes']
                    assert int(Policy.BUILTIN_POLICIES[policy_name]['dh_modulus_sizes']['diffie-hellman-group-exchange-sha256']) >= 2048

            else:  # Client-specific tests.

                # These must be set to None for client policies, since they have no meaning otherwise.
                assert Policy.BUILTIN_POLICIES[policy_name]['hostkey_sizes'] is None
                assert Policy.BUILTIN_POLICIES[policy_name]['dh_modulus_sizes'] is None

            # Ensure that each built-in policy can be loaded with Policy.load_builtin_policy().
            assert Policy.load_builtin_policy(policy_name) is not None

        # Ensure that both server and client policy names are returned.
        server_policy_names, client_policy_names = Policy.list_builtin_policies()
        assert len(server_policy_names) > 0
        assert len(client_policy_names) > 0


    def test_policy_basic(self):
        '''Ensure that a basic policy can be parsed correctly.'''

        policy_data = '''# This is a comment
name = "Test Policy"
version = 1

compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        assert str(policy) == "Name: [Test Policy]\nVersion: [1]\nBanner: {undefined}\nCompressions: comp_alg1\nHost Keys: key_alg1\nOptional Host Keys: {undefined}\nKey Exchanges: kex_alg1, kex_alg2\nCiphers: cipher_alg1, cipher_alg2, cipher_alg3\nMACs: mac_alg1, mac_alg2, mac_alg3\nHost Key Sizes: {undefined}\nDH Modulus Sizes: {undefined}\nServer Policy: True"


    def test_policy_invalid_1(self):
        '''Basic policy, but with 'ciphersx' instead of 'ciphers'.'''

        policy_data = '''# This is a comment
name = "Test Policy"
version = 1

compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphersx = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        failed = False
        try:
            self.Policy(policy_data=policy_data)
        except ValueError:
            failed = True

        assert failed, "Invalid policy did not cause Policy object to throw exception"


    def test_policy_invalid_2(self):
        '''Basic policy, but is missing the required name field.'''

        policy_data = '''# This is a comment
#name = "Test Policy"
version = 1

compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        failed = False
        try:
            self.Policy(policy_data=policy_data)
        except ValueError:
            failed = True

        assert failed, "Invalid policy did not cause Policy object to throw exception"


    def test_policy_invalid_3(self):
        '''Basic policy, but is missing the required version field.'''

        policy_data = '''# This is a comment
name = "Test Policy"
#version = 1

compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        failed = False
        try:
            self.Policy(policy_data=policy_data)
        except ValueError:
            failed = True

        assert failed, "Invalid policy did not cause Policy object to throw exception"


    def test_policy_invalid_4(self):
        '''Basic policy, but is missing quotes in the name field.'''

        policy_data = '''# This is a comment
name = Test Policy
version = 1

compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        failed = False
        try:
            self.Policy(policy_data=policy_data)
        except ValueError:
            failed = True

        assert failed, "Invalid policy did not cause Policy object to throw exception"


    def test_policy_invalid_5(self):
        '''Basic policy, but is missing quotes in the banner field.'''

        policy_data = '''# This is a comment
name = "Test Policy"
version = 1

banner = 0mg
compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        failed = False
        try:
            self.Policy(policy_data=policy_data)
        except ValueError:
            failed = True

        assert failed, "Invalid policy did not cause Policy object to throw exception"


    def test_policy_invalid_6(self):
        '''Basic policy, but is missing quotes in the header field.'''

        policy_data = '''# This is a comment
name = "Test Policy"
version = 1

header = 0mg
compressions = comp_alg1
host keys = key_alg1
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        failed = False
        try:
            self.Policy(policy_data=policy_data)
        except ValueError:
            failed = True

        assert failed, "Invalid policy did not cause Policy object to throw exception"


    def test_policy_create_1(self):
        '''Creates a policy from a kex and ensures it is generated exactly as expected.'''

        kex = self._get_kex()
        pol_data = self.Policy.create('www.l0l.com', 'bannerX', kex, False)

        # Today's date is embedded in the policy, so filter it out to get repeatable results.
        pol_data = pol_data.replace(date.today().strftime('%Y/%m/%d'), '[todays date]')

        # Instead of writing out the entire expected policy--line by line--just check that it has the expected hash.
        assert hashlib.sha256(pol_data.encode('ascii')).hexdigest() == '4af7777fb57a1dad0cf438c899a11d4f625fd9276ea3bb5ef5c9fe8806cb47dc'


    def test_policy_evaluate_passing_1(self):
        '''Creates a policy and evaluates it against the same server'''

        kex = self._get_kex()
        policy_data = self.Policy.create('www.l0l.com', None, kex, False)
        policy = self.Policy(policy_data=policy_data)

        ret, errors, error_str = policy.evaluate('SSH Server 1.0', kex)
        assert ret is True
        assert len(errors) == 0
        print(error_str)
        assert len(error_str) == 0


    def test_policy_evaluate_failing_1(self):
        '''Ensure that a policy with a specified banner fails against a server with a different banner'''

        policy_data = '''name = "Test Policy"
version = 1
banner = "XXX mismatched banner XXX"
compressions = comp_alg1, comp_alg2
host keys = key_alg1, key_alg2
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 1
        assert error_str.find('Banner did not match.') != -1


    def test_policy_evaluate_failing_2(self):
        '''Ensure that a mismatched compressions list results in a failure'''

        policy_data = '''name = "Test Policy"
version = 1
compressions = XXXmismatchedXXX, comp_alg1, comp_alg2
host keys = key_alg1, key_alg2
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 1
        assert error_str.find('Compression did not match.') != -1


    def test_policy_evaluate_failing_3(self):
        '''Ensure that a mismatched host keys results in a failure'''

        policy_data = '''name = "Test Policy"
version = 1
compressions = comp_alg1, comp_alg2
host keys = XXXmismatchedXXX, key_alg1, key_alg2
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 1
        assert error_str.find('Host keys did not match.') != -1


    def test_policy_evaluate_failing_4(self):
        '''Ensure that a mismatched key exchange list results in a failure'''

        policy_data = '''name = "Test Policy"
version = 1
compressions = comp_alg1, comp_alg2
host keys = key_alg1, key_alg2
key exchanges = XXXmismatchedXXX, kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 1
        assert error_str.find('Key exchanges did not match.') != -1


    def test_policy_evaluate_failing_5(self):
        '''Ensure that a mismatched cipher list results in a failure'''

        policy_data = '''name = "Test Policy"
version = 1
compressions = comp_alg1, comp_alg2
host keys = key_alg1, key_alg2
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, XXXmismatched, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 1
        assert error_str.find('Ciphers did not match.') != -1


    def test_policy_evaluate_failing_6(self):
        '''Ensure that a mismatched MAC list results in a failure'''

        policy_data = '''name = "Test Policy"
version = 1
compressions = comp_alg1, comp_alg2
host keys = key_alg1, key_alg2
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, XXXmismatched, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 1
        assert error_str.find('MACs did not match.') != -1


    def test_policy_evaluate_failing_7(self):
        '''Ensure that a mismatched host keys and MACs results in a failure'''

        policy_data = '''name = "Test Policy"
version = 1
compressions = comp_alg1, comp_alg2
host keys = key_alg1, key_alg2, XXXmismatchedXXX
key exchanges = kex_alg1, kex_alg2
ciphers = cipher_alg1, cipher_alg2, cipher_alg3
macs = mac_alg1, mac_alg2, XXXmismatchedXXX, mac_alg3'''

        policy = self.Policy(policy_data=policy_data)
        ret, errors, error_str = policy.evaluate('SSH Server 1.0', self._get_kex())
        assert ret is False
        assert len(errors) == 2
        assert error_str.find('Host keys did not match.') != -1
        assert error_str.find('MACs did not match.') != -1



import socket
import pytest


# pylint: disable=attribute-defined-outside-init,protected-access
class TestResolve:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.AuditConf = ssh_audit.AuditConf
        self.audit = ssh_audit.audit
        self.OutputBuffer = ssh_audit.OutputBuffer
        self.ssh_socket = ssh_audit.SSH_Socket

    def _conf(self):
        conf = self.AuditConf('localhost', 22)
        conf.colors = False
        conf.batch = True
        return conf

    def test_resolve_error(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.gsock.addrinfodata['localhost#22'] = socket.gaierror(8, 'hostname nor servname provided, or not known')
        conf = self._conf()
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        # output_spy.begin()
        with pytest.raises(socket.gaierror):
            list(s._resolve())
        # lines = output_spy.flush()
        # assert len(lines) == 1
        # assert 'hostname nor servname provided' in lines[-1]

    def test_resolve_hostname_without_records(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.gsock.addrinfodata['localhost#22'] = []
        conf = self._conf()
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        output_spy.begin()
        r = list(s._resolve())
        assert len(r) == 0

    def test_resolve_ipv4(self, virtual_socket):
        conf = self._conf()
        conf.ipv4 = True
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        r = list(s._resolve())
        assert len(r) == 1
        assert r[0] == (socket.AF_INET, ('127.0.0.1', 22))

    def test_resolve_ipv6(self, virtual_socket):
        conf = self._conf()
        conf.ipv6 = True
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        r = list(s._resolve())
        assert len(r) == 1
        assert r[0] == (socket.AF_INET6, ('::1', 22))

    def test_resolve_ipv46_both(self, virtual_socket):
        conf = self._conf()
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        r = list(s._resolve())
        assert len(r) == 2
        assert r[0] == (socket.AF_INET, ('127.0.0.1', 22))
        assert r[1] == (socket.AF_INET6, ('::1', 22))

    def test_resolve_ipv46_order(self, virtual_socket):
        conf = self._conf()
        conf.ipv4 = True
        conf.ipv6 = True
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        r = list(s._resolve())
        assert len(r) == 2
        assert r[0] == (socket.AF_INET, ('127.0.0.1', 22))
        assert r[1] == (socket.AF_INET6, ('::1', 22))
        conf = self._conf()
        conf.ipv6 = True
        conf.ipv4 = True
        s = self.ssh_socket(self.OutputBuffer(), 'localhost', 22, conf.ip_version_preference)
        r = list(s._resolve())
        assert len(r) == 2
        assert r[0] == (socket.AF_INET6, ('::1', 22))
        assert r[1] == (socket.AF_INET, ('127.0.0.1', 22))



import pytest


# pylint: disable=attribute-defined-outside-init
class TestOutputBuffer:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.OutputBuffer = ssh_audit.OutputBuffer

    def test_outputbuffer_no_lines(self, output_spy):
        output_spy.begin()
        obuf = self.OutputBuffer()
        obuf.write()
        assert output_spy.flush() == ['']
        output_spy.begin()

    def test_outputbuffer_defaults(self):
        obuf = self.OutputBuffer()
        # default: on
        assert obuf.batch is False
        assert obuf.use_colors is True
        assert obuf.level == 'info'

    def test_outputbuffer_colors(self, output_spy):
        out = self.OutputBuffer()

        # Test without colors.
        out.use_colors = False

        output_spy.begin()
        out.info('info color')
        out.write()
        assert output_spy.flush() == ['info color']

        output_spy.begin()
        out.head('head color')
        out.write()
        assert output_spy.flush() == ['head color']

        output_spy.begin()
        out.good('good color')
        out.write()
        assert output_spy.flush() == ['good color']

        output_spy.begin()
        out.warn('warn color')
        out.write()
        assert output_spy.flush() == ['warn color']

        output_spy.begin()
        out.fail('fail color')
        out.write()
        assert output_spy.flush() == ['fail color']

        # If colors aren't supported by this system, skip the color tests.
        if not out.colors_supported:
            return

        # Test with colors.
        out.use_colors = True

        output_spy.begin()
        out.info('info color')
        out.write()
        assert output_spy.flush() == ['info color']

        output_spy.begin()
        out.head('head color')
        out.write()
        assert output_spy.flush() in [['\x1b[0;36mhead color\x1b[0m'], ['\x1b[0;96mhead color\x1b[0m']]

        output_spy.begin()
        out.good('good color')
        out.write()
        assert output_spy.flush() in [['\x1b[0;32mgood color\x1b[0m'], ['\x1b[0;92mgood color\x1b[0m']]

        output_spy.begin()
        out.warn('warn color')
        out.write()
        assert output_spy.flush() in [['\x1b[0;33mwarn color\x1b[0m'], ['\x1b[0;93mwarn color\x1b[0m']]

        output_spy.begin()
        out.fail('fail color')
        out.write()
        assert output_spy.flush() in [['\x1b[0;31mfail color\x1b[0m'], ['\x1b[0;91mfail color\x1b[0m']]

    def test_outputbuffer_sep(self, output_spy):
        out = self.OutputBuffer()
        output_spy.begin()
        out.sep()
        out.sep()
        out.sep()
        out.write()
        assert output_spy.flush() == ['', '', '']

    def test_outputbuffer_levels(self):
        out = self.OutputBuffer()
        assert out.get_level('info') == 0
        assert out.get_level('good') == 0
        assert out.get_level('warn') == 1
        assert out.get_level('fail') == 2
        assert out.get_level('unknown') > 2

    def test_outputbuffer_level_property(self):
        out = self.OutputBuffer()
        out.level = 'info'
        assert out.level == 'info'
        out.level = 'good'
        assert out.level == 'info'
        out.level = 'warn'
        assert out.level == 'warn'
        out.level = 'fail'
        assert out.level == 'fail'
        out.level = 'invalid level'
        assert out.level == 'unknown'

    def test_outputbuffer_level(self, output_spy):
        out = self.OutputBuffer()
        # visible: all
        out.level = 'info'
        output_spy.begin()
        out.info('info color')
        out.head('head color')
        out.good('good color')
        out.warn('warn color')
        out.fail('fail color')
        out.write()
        assert len(output_spy.flush()) == 5
        # visible: head, warn, fail
        out.level = 'warn'
        output_spy.begin()
        out.info('info color')
        out.head('head color')
        out.good('good color')
        out.warn('warn color')
        out.fail('fail color')
        out.write()
        assert len(output_spy.flush()) == 3
        # visible: head, fail
        out.level = 'fail'
        output_spy.begin()
        out.info('info color')
        out.head('head color')
        out.good('good color')
        out.warn('warn color')
        out.fail('fail color')
        out.write()
        assert len(output_spy.flush()) == 2
        # visible: head
        out.level = 'invalid level'
        output_spy.begin()
        out.info('info color')
        out.head('head color')
        out.good('good color')
        out.warn('warn color')
        out.fail('fail color')
        out.write()
        assert len(output_spy.flush()) == 1

    def test_outputbuffer_batch(self, output_spy):
        out = self.OutputBuffer()
        # visible: all
        output_spy.begin()
        out.level = 'info'
        out.batch = False
        out.info('info color')
        out.head('head color')
        out.good('good color')
        out.warn('warn color')
        out.fail('fail color')
        out.write()
        assert len(output_spy.flush()) == 5
        # visible: all except head
        output_spy.begin()
        out.level = 'info'
        out.batch = True
        out.info('info color')
        out.head('head color')
        out.good('good color')
        out.warn('warn color')
        out.fail('fail color')
        out.write()
        assert len(output_spy.flush()) == 4



import pytest

from ssh_audit.ssh2_kexdb import SSH2_KexDB


class Test_SSH2_KexDB:

    @pytest.fixture(autouse=True)
    def init(self):
        self.db = SSH2_KexDB.get_db()

    def test_ssh2_kexdb(self):
        '''Ensures that the SSH2_KexDB.ALGORITHMS dictionary is in the right format.'''

        db_keys = list(self.db.keys())
        db_keys.sort()

        # Ensure only these keys exist in the database.
        assert db_keys == ['enc', 'kex', 'key', 'mac']

        # For 'enc', 'kex', etc...
        for alg_type in self.db:

            # Iterate over algorithms within this type (i.e.: all 'enc' algorithms, all 'kex' algorithms, etc).
            for alg_name in self.db[alg_type]:

                # Get the list of failures, warnings, etc., for this algorithm.
                alg_data = self.db[alg_type][alg_name]

                # This list must be between 1 and 4 entries long.
                assert 1 <= len(alg_data) <= 4

                # The first entry denotes the versions when this algorithm was added to OpenSSH, Dropbear, and/or libssh, followed by when it was deprecated, and finally when it was removed.  Hence it must have between 0 and 3 entries.
                added_entry = alg_data[0]
                assert 0 <= len(added_entry) <= 3



import pytest
from ssh_audit.ssh_audit import process_commandline


# pylint: disable=attribute-defined-outside-init
class TestAuditConf:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.AuditConf = ssh_audit.AuditConf
        self.OutputBuffer = ssh_audit.OutputBuffer()
        self.usage = ssh_audit.usage
        self.process_commandline = process_commandline

    @staticmethod
    def _test_conf(conf, **kwargs):
        options = {
            'host': '',
            'port': 22,
            'ssh1': True,
            'ssh2': True,
            'batch': False,
            'colors': True,
            'verbose': False,
            'level': 'info',
            'ipv4': False,
            'ipv6': False
        }
        for k, v in kwargs.items():
            options[k] = v
        assert conf.host == options['host']
        assert conf.port == options['port']
        assert conf.ssh1 is options['ssh1']
        assert conf.ssh2 is options['ssh2']
        assert conf.batch is options['batch']
        assert conf.colors is options['colors']
        assert conf.verbose is options['verbose']
        assert conf.level == options['level']
        assert conf.ipv4 == options['ipv4']
        assert conf.ipv6 == options['ipv6']

    def test_audit_conf_defaults(self):
        conf = self.AuditConf()
        self._test_conf(conf)

    def test_audit_conf_booleans(self):
        conf = self.AuditConf()
        for p in ['ssh1', 'ssh2', 'batch', 'colors', 'verbose']:
            for v in [True, 1]:
                setattr(conf, p, v)
                assert getattr(conf, p) is True
            for v in [False, 0]:
                setattr(conf, p, v)
                assert getattr(conf, p) is False

    def test_audit_conf_port(self):
        conf = self.AuditConf()
        for port in [22, 2222]:
            conf.port = port
            assert conf.port == port
        for port in [-1, 0, 65536, 99999]:
            with pytest.raises(ValueError) as excinfo:
                conf.port = port
            excinfo.match(r'.*invalid port.*')

    def test_audit_conf_ip_version_preference(self):
        # ipv4-only
        conf = self.AuditConf()
        conf.ipv4 = True
        assert conf.ipv4 is True
        assert conf.ipv6 is False
        assert conf.ip_version_preference == [4]
        # ipv6-only
        conf = self.AuditConf()
        conf.ipv6 = True
        assert conf.ipv4 is False
        assert conf.ipv6 is True
        assert conf.ip_version_preference == [6]
        # ipv4-preferred
        conf = self.AuditConf()
        conf.ipv4 = True
        conf.ipv6 = True
        assert conf.ipv4 is True
        assert conf.ipv6 is True
        assert conf.ip_version_preference == [4, 6]
        # ipv6-preferred
        conf = self.AuditConf()
        conf.ipv6 = True
        conf.ipv4 = True
        assert conf.ipv4 is True
        assert conf.ipv6 is True
        assert conf.ip_version_preference == [6, 4]
        # defaults
        conf = self.AuditConf()
        assert conf.ipv4 is False
        assert conf.ipv6 is False
        assert conf.ip_version_preference == []

    def test_audit_conf_level(self):
        conf = self.AuditConf()
        for level in ['info', 'warn', 'fail']:
            conf.level = level
            assert conf.level == level
        for level in ['head', 'good', 'unknown', None]:
            with pytest.raises(ValueError) as excinfo:
                conf.level = level
            excinfo.match(r'.*invalid level.*')

    def test_audit_conf_process_commandline(self):
        # pylint: disable=too-many-statements
        c = lambda x: self.process_commandline(self.OutputBuffer, x.split(), self.usage)  # noqa
        with pytest.raises(SystemExit):
            conf = c('')
        with pytest.raises(SystemExit):
            conf = c('-x')
        with pytest.raises(SystemExit):
            conf = c('-h')
        with pytest.raises(SystemExit):
            conf = c('--help')
        with pytest.raises(SystemExit):
            conf = c(':')
        with pytest.raises(SystemExit):
            conf = c(':22')
        conf = c('localhost')
        self._test_conf(conf, host='localhost')
        conf = c('github.com')
        self._test_conf(conf, host='github.com')
        conf = c('localhost:2222')
        self._test_conf(conf, host='localhost', port=2222)
        conf = c('-p 2222 localhost')
        self._test_conf(conf, host='localhost', port=2222)
        conf = c('2001:4860:4860::8888')
        self._test_conf(conf, host='2001:4860:4860::8888')
        conf = c('[2001:4860:4860::8888]:22')
        self._test_conf(conf, host='2001:4860:4860::8888')
        conf = c('[2001:4860:4860::8888]:2222')
        self._test_conf(conf, host='2001:4860:4860::8888', port=2222)
        conf = c('-p 2222 2001:4860:4860::8888')
        self._test_conf(conf, host='2001:4860:4860::8888', port=2222)
        with pytest.raises(ValueError):
            conf = c('localhost:abc')
        with pytest.raises(SystemExit):
            conf = c('-p abc localhost')
        with pytest.raises(ValueError):
            conf = c('localhost:-22')
        with pytest.raises(SystemExit):
            conf = c('-p -22 localhost')
        with pytest.raises(ValueError):
            conf = c('localhost:99999')
        with pytest.raises(SystemExit):
            conf = c('-p 99999 localhost')
        conf = c('-1 localhost')
        self._test_conf(conf, host='localhost', ssh1=True, ssh2=False)
        conf = c('-2 localhost')
        self._test_conf(conf, host='localhost', ssh1=False, ssh2=True)
        conf = c('-12 localhost')
        self._test_conf(conf, host='localhost', ssh1=True, ssh2=True)
        conf = c('-4 localhost')
        self._test_conf(conf, host='localhost', ipv4=True, ipv6=False, ipvo=(4,))
        conf = c('-6 localhost')
        self._test_conf(conf, host='localhost', ipv4=False, ipv6=True, ipvo=(6,))
        conf = c('-46 localhost')
        self._test_conf(conf, host='localhost', ipv4=True, ipv6=True, ipvo=(4, 6))
        conf = c('-64 localhost')
        self._test_conf(conf, host='localhost', ipv4=True, ipv6=True, ipvo=(6, 4))
        conf = c('-b localhost')
        self._test_conf(conf, host='localhost', batch=True, verbose=True)
        conf = c('-n localhost')
        self._test_conf(conf, host='localhost', colors=False)
        conf = c('-v localhost')
        self._test_conf(conf, host='localhost', verbose=True)
        conf = c('-l info localhost')
        self._test_conf(conf, host='localhost', level='info')
        conf = c('-l warn localhost')
        self._test_conf(conf, host='localhost', level='warn')
        conf = c('-l fail localhost')
        self._test_conf(conf, host='localhost', level='fail')
        with pytest.raises(SystemExit):
            conf = c('-l something localhost')



import struct
import pytest

from ssh_audit.auditconf import AuditConf
from ssh_audit.fingerprint import Fingerprint
from ssh_audit.outputbuffer import OutputBuffer
from ssh_audit.protocol import Protocol
from ssh_audit.readbuf import ReadBuf
from ssh_audit.ssh1 import SSH1
from ssh_audit.ssh1_publickeymessage import SSH1_PublicKeyMessage
from ssh_audit.ssh_audit import audit
from ssh_audit.writebuf import WriteBuf


# pylint: disable=line-too-long,attribute-defined-outside-init
class TestSSH1:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.OutputBuffer = OutputBuffer
        self.protocol = Protocol
        self.ssh1 = SSH1
        self.PublicKeyMessage = SSH1_PublicKeyMessage
        self.rbuf = ReadBuf
        self.wbuf = WriteBuf
        self.audit = audit
        self.AuditConf = AuditConf
        self.fingerprint = Fingerprint

    def _conf(self):
        conf = self.AuditConf('localhost', 22)
        conf.colors = False
        conf.batch = True
        conf.verbose = True
        conf.ssh1 = True
        conf.ssh2 = False
        return conf

    def _create_ssh1_packet(self, payload, valid_crc=True):
        padding = -(len(payload) + 4) % 8
        plen = len(payload) + 4
        pad_bytes = b'\x00' * padding
        cksum = self.ssh1.crc32(pad_bytes + payload) if valid_crc else 0
        data = struct.pack('>I', plen) + pad_bytes + payload + struct.pack('>I', cksum)
        return data

    @classmethod
    def _server_key(cls):
        return (1024, 0x10001, 0xee6552da432e0ac2c422df1a51287507748bfe3b5e3e4fa989a8f49fdc163a17754939ef18ef8a667ea3b71036a151fcd7f5e01ceef1e4439864baf3ac569047582c69d6c128212e0980dcb3168f00d371004039983f6033cd785b8b8f85096c7d9405cbfdc664e27c966356a6b4eb6ee20ad43414b50de18b22829c1880b551)

    @classmethod
    def _host_key(cls):
        return (2048, 0x10001, 0xdfa20cd2a530ccc8c870aa60d9feb3b35deeab81c3215a96557abbd683d21f4600f38e475d87100da9a4404220eeb3bb5584e5a2b5b48ffda58530ea19104a32577d7459d91e76aa711b241050f4cc6d5327ccce254f371acad3be56d46eb5919b73f20dbdb1177b700f00891c5bf4ed128bb90ed541b778288285bcfa28432ab5cbcb8321b6e24760e998e0daa519f093a631e44276d7dd252ce0c08c75e2ab28a7349ead779f97d0f20a6d413bf3623cd216dc35375f6366690bcc41e3b2d5465840ec7ee0dc7e3f1c101d674a0c7dbccbc3942788b111396add2f8153b46a0e4b50d66e57ee92958f1c860dd97cc0e40e32febff915343ed53573142bdf4b)

    def _pkm_payload(self):
        w = self.wbuf()
        w.write(b'\x88\x99\xaa\xbb\xcc\xdd\xee\xff')
        b, e, m = self._server_key()
        w.write_int(b).write_mpint1(e).write_mpint1(m)
        b, e, m = self._host_key()
        w.write_int(b).write_mpint1(e).write_mpint1(m)
        w.write_int(2)
        w.write_int(72)
        w.write_int(36)
        return w.write_flush()

    def test_crc32(self):
        assert self.ssh1.crc32(b'') == 0x00
        assert self.ssh1.crc32(b'The quick brown fox jumps over the lazy dog') == 0xb9c60808

    def test_fingerprint(self):
        # pylint: disable=protected-access
        b, e, m = self._host_key()
        fpd = self.wbuf._create_mpint(m, False)
        fpd += self.wbuf._create_mpint(e, False)
        fp = self.fingerprint(fpd)
        assert b == 2048
        assert fp.md5 == 'MD5:9d:26:f8:39:fc:20:9d:9b:ca:cc:4a:0f:e1:93:f5:96'
        assert fp.sha256 == 'SHA256:vZdx3mhzbvVJmn08t/ruv8WDhJ9jfKYsCTuSzot+QIs'

    def _assert_pkm_keys(self, pkm, skey, hkey):
        b, e, m = skey
        assert pkm.server_key_bits == b
        assert pkm.server_key_public_exponent == e
        assert pkm.server_key_public_modulus == m
        b, e, m = hkey
        assert pkm.host_key_bits == b
        assert pkm.host_key_public_exponent == e
        assert pkm.host_key_public_modulus == m

    def _assert_pkm_fields(self, pkm, skey, hkey):
        assert pkm is not None
        assert pkm.cookie == b'\x88\x99\xaa\xbb\xcc\xdd\xee\xff'
        self._assert_pkm_keys(pkm, skey, hkey)
        assert pkm.protocol_flags == 2
        assert pkm.supported_ciphers_mask == 72
        assert pkm.supported_ciphers == ['3des', 'blowfish']
        assert pkm.supported_authentications_mask == 36
        assert pkm.supported_authentications == ['rsa', 'tis']
        fp = self.fingerprint(pkm.host_key_fingerprint_data)
        assert fp.md5 == 'MD5:9d:26:f8:39:fc:20:9d:9b:ca:cc:4a:0f:e1:93:f5:96'
        assert fp.sha256 == 'SHA256:vZdx3mhzbvVJmn08t/ruv8WDhJ9jfKYsCTuSzot+QIs'

    def test_pkm_init(self):
        cookie = b'\x88\x99\xaa\xbb\xcc\xdd\xee\xff'
        pflags, cmask, amask = 2, 72, 36
        skey, hkey = self._server_key(), self._host_key()
        pkm = self.PublicKeyMessage(cookie, skey, hkey, pflags, cmask, amask)
        self._assert_pkm_fields(pkm, skey, hkey)
        for skey2 in ([], [0], [0, 1], [0, 1, 2, 3]):
            with pytest.raises(ValueError):
                pkm = self.PublicKeyMessage(cookie, skey2, hkey, pflags, cmask, amask)
        for hkey2 in ([], [0], [0, 1], [0, 1, 2, 3]):
            with pytest.raises(ValueError):
                print(hkey2)
                pkm = self.PublicKeyMessage(cookie, skey, hkey2, pflags, cmask, amask)

    def test_pkm_read(self):
        pkm = self.PublicKeyMessage.parse(self._pkm_payload())
        self._assert_pkm_fields(pkm, self._server_key(), self._host_key())

    def test_pkm_payload(self):
        cookie = b'\x88\x99\xaa\xbb\xcc\xdd\xee\xff'
        skey, hkey = self._server_key(), self._host_key()
        pflags, cmask, amask = 2, 72, 36
        pkm1 = self.PublicKeyMessage(cookie, skey, hkey, pflags, cmask, amask)
        pkm2 = self.PublicKeyMessage.parse(self._pkm_payload())
        assert pkm1.payload == pkm2.payload

    def test_ssh1_server_simple(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        w = self.wbuf()
        w.write_byte(self.protocol.SMSG_PUBLIC_KEY)
        w.write(self._pkm_payload())
        vsocket.rdata.append(b'SSH-1.5-OpenSSH_7.2 ssh-audit-test\r\n')
        vsocket.rdata.append(self._create_ssh1_packet(w.write_flush()))
        output_spy.begin()
        out = self.OutputBuffer()
        self.audit(out, self._conf())
        out.write()
        lines = output_spy.flush()
        assert len(lines) == 21

    def test_ssh1_server_invalid_first_packet(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        w = self.wbuf()
        w.write_byte(self.protocol.SMSG_PUBLIC_KEY + 1)
        w.write(self._pkm_payload())
        vsocket.rdata.append(b'SSH-1.5-OpenSSH_7.2 ssh-audit-test\r\n')
        vsocket.rdata.append(self._create_ssh1_packet(w.write_flush()))
        output_spy.begin()
        out = self.OutputBuffer()
        ret = self.audit(out, self._conf())
        out.write()
        assert ret != 0
        lines = output_spy.flush()
        assert len(lines) == 14
        assert 'unknown message' in lines[-1]

    def test_ssh1_server_invalid_checksum(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        w = self.wbuf()
        w.write_byte(self.protocol.SMSG_PUBLIC_KEY + 1)
        w.write(self._pkm_payload())
        vsocket.rdata.append(b'SSH-1.5-OpenSSH_7.2 ssh-audit-test\r\n')
        vsocket.rdata.append(self._create_ssh1_packet(w.write_flush(), False))
        output_spy.begin()
        out = self.OutputBuffer()
        with pytest.raises(SystemExit):
            self.audit(out, self._conf())
        out.write()
        lines = output_spy.flush()
        assert len(lines) == 3
        assert ('checksum' in lines[0]) or ('checksum' in lines[1]) or ('checksum' in lines[2])



"""
   The MIT License (MIT)

   Copyright (C) 2017-2021 Joe Testa (jtesta@positronsecurity.com)
   Copyright (C) 2017 Andris Raugulis (moo@arthepsy.eu)

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
"""
# pylint: disable=unused-import
from typing import Dict, List, Set, Sequence, Tuple, Iterable  # noqa: F401
from typing import Callable, Optional, Union, Any  # noqa: F401

from ssh_audit.policy import Policy
from ssh_audit.utils import Utils


class AuditConf:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, host: str = '', port: int = 22) -> None:
        self.host = host
        self.port = port
        self.ssh1 = True
        self.ssh2 = True
        self.batch = False
        self.client_audit = False
        self.colors = True
        self.json = False
        self.json_print_indent = False
        self.verbose = False
        self.level = 'info'
        self.ip_version_preference: List[int] = []  # Holds only 5 possible values: [] (no preference), [4] (use IPv4 only), [6] (use IPv6 only), [46] (use both IPv4 and IPv6, but prioritize v4), and [64] (use both IPv4 and IPv6, but prioritize v6).
        self.ipv4 = False
        self.ipv6 = False
        self.make_policy = False  # When True, creates a policy file from an audit scan.
        self.policy_file: Optional[str] = None   # File system path to a policy
        self.policy: Optional[Policy] = None  # Policy object
        self.timeout = 5.0
        self.timeout_set = False  # Set to True when the user explicitly sets it.
        self.target_file: Optional[str] = None
        self.target_list: List[str] = []
        self.threads = 32
        self.list_policies = False
        self.lookup = ''
        self.manual = False
        self.debug = False
        self.gex_test = ''

    def __setattr__(self, name: str, value: Union[str, int, float, bool, Sequence[int]]) -> None:
        valid = False
        if name in ['batch', 'client_audit', 'colors', 'json', 'json_print_indent', 'list_policies', 'manual', 'make_policy', 'ssh1', 'ssh2', 'timeout_set', 'verbose', 'debug']:
            valid, value = True, bool(value)
        elif name in ['ipv4', 'ipv6']:
            valid, value = True, bool(value)
            if len(self.ip_version_preference) == 2:  # Being called more than twice is not valid.
                valid = False
            elif value:
                self.ip_version_preference.append(4 if name == 'ipv4' else 6)
        elif name == 'port':
            valid, port = True, Utils.parse_int(value)
            if port < 1 or port > 65535:
                raise ValueError('invalid port: {}'.format(value))
            value = port
        elif name in ['level']:
            if value not in ('info', 'warn', 'fail'):
                raise ValueError('invalid level: {}'.format(value))
            valid = True
        elif name == 'host':
            valid = True
        elif name == 'timeout':
            value = Utils.parse_float(value)
            if value == -1.0:
                raise ValueError('invalid timeout: {}'.format(value))
            valid = True
        elif name in ['ip_version_preference', 'lookup', 'policy_file', 'policy', 'target_file', 'target_list', 'gex_test']:
            valid = True
        elif name == "threads":
            valid, num_threads = True, Utils.parse_int(value)
            if num_threads < 1:
                raise ValueError('invalid number of threads: {}'.format(value))
            value = num_threads

        if valid:
            object.__setattr__(self, name, value)



import pytest

from ssh_audit.outputbuffer import OutputBuffer
from ssh_audit.ssh_socket import SSH_Socket


# pylint: disable=attribute-defined-outside-init
class TestSocket:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.OutputBuffer = OutputBuffer
        self.ssh_socket = SSH_Socket

    def test_invalid_host(self, virtual_socket):
        with pytest.raises(ValueError):
            self.ssh_socket(self.OutputBuffer(), None, 22)

    def test_invalid_port(self, virtual_socket):
        with pytest.raises(ValueError):
            self.ssh_socket(self.OutputBuffer(), 'localhost', 'abc')
        with pytest.raises(ValueError):
            self.ssh_socket(self.OutputBuffer(), 'localhost', -1)
        with pytest.raises(ValueError):
            self.ssh_socket(self.OutputBuffer(), 'localhost', 0)
        with pytest.raises(ValueError):
            self.ssh_socket(self.OutputBuffer(), 'localhost', 65536)

    def test_not_connected_socket(self, virtual_socket):
        sock = self.ssh_socket(self.OutputBuffer(), 'localhost', 22)
        banner, header, err = sock.get_banner()
        assert banner is None
        assert len(header) == 0
        assert err == 'not connected'
        s, e = sock.recv()
        assert s == -1
        assert e == 'not connected'
        s, e = sock.send('nothing')
        assert s == -1
        assert e == 'not connected'
        s, e = sock.send_packet()
        assert s == -1
        assert e == 'not connected'



import socket
import errno
import pytest

from ssh_audit.outputbuffer import OutputBuffer


# pylint: disable=attribute-defined-outside-init
class TestErrors:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.AuditConf = ssh_audit.AuditConf
        self.OutputBuffer = ssh_audit.OutputBuffer
        self.audit = ssh_audit.audit

    def _conf(self):
        conf = self.AuditConf('localhost', 22)
        conf.colors = False
        conf.batch = True
        return conf

    def _audit(self, spy, conf=None, exit_expected=False):
        if conf is None:
            conf = self._conf()
        spy.begin()

        out = OutputBuffer()
        if exit_expected:
            with pytest.raises(SystemExit):
                self.audit(out, conf)
        else:
            ret = self.audit(out, conf)
            assert ret != 0

        out.write()
        lines = spy.flush()

        # If the last line is empty, delete it.
        if len(lines) > 1 and lines[-1] == '':
            del lines[-1]

        return lines

    def test_connection_unresolved(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.gsock.addrinfodata['localhost#22'] = []
        lines = self._audit(output_spy, exit_expected=True)
        assert len(lines) == 1
        assert 'has no DNS records' in lines[-1]

    def test_connection_refused(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.errors['connect'] = socket.error(errno.ECONNREFUSED, 'Connection refused')
        lines = self._audit(output_spy, exit_expected=True)
        assert len(lines) == 1
        assert 'Connection refused' in lines[-1]

    def test_connection_timeout(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.errors['connect'] = socket.timeout('timed out')
        lines = self._audit(output_spy, exit_expected=True)
        assert len(lines) == 1
        assert 'timed out' in lines[-1]

    def test_recv_empty(self, output_spy, virtual_socket):
        lines = self._audit(output_spy)
        assert len(lines) == 1
        assert 'did not receive banner' in lines[-1]

    def test_recv_timeout(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(socket.timeout('timed out'))
        lines = self._audit(output_spy)
        assert len(lines) == 1
        assert 'did not receive banner' in lines[-1]
        assert 'timed out' in lines[-1]

    def test_recv_retry_till_timeout(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(socket.error(errno.EAGAIN, 'Resource temporarily unavailable'))
        vsocket.rdata.append(socket.error(errno.EWOULDBLOCK, 'Resource temporarily unavailable'))
        vsocket.rdata.append(socket.error(errno.EAGAIN, 'Resource temporarily unavailable'))
        vsocket.rdata.append(socket.timeout('timed out'))
        lines = self._audit(output_spy)
        assert len(lines) == 1
        assert 'did not receive banner' in lines[-1]
        assert 'timed out' in lines[-1]

    def test_recv_retry_till_reset(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(socket.error(errno.EAGAIN, 'Resource temporarily unavailable'))
        vsocket.rdata.append(socket.error(errno.EWOULDBLOCK, 'Resource temporarily unavailable'))
        vsocket.rdata.append(socket.error(errno.EAGAIN, 'Resource temporarily unavailable'))
        vsocket.rdata.append(socket.error(errno.ECONNRESET, 'Connection reset by peer'))
        lines = self._audit(output_spy)
        assert len(lines) == 1
        assert 'did not receive banner' in lines[-1]
        assert 'reset by peer' in lines[-1]

    def test_connection_closed_before_banner(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(socket.error(errno.ECONNRESET, 'Connection reset by peer'))
        lines = self._audit(output_spy)
        assert len(lines) == 1
        assert 'did not receive banner' in lines[-1]
        assert 'reset by peer' in lines[-1]

    def test_connection_closed_after_header(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'header line 1\n')
        vsocket.rdata.append(b'\n')
        vsocket.rdata.append(b'header line 2\n')
        vsocket.rdata.append(socket.error(errno.ECONNRESET, 'Connection reset by peer'))
        lines = self._audit(output_spy)
        assert len(lines) == 3
        assert 'did not receive banner' in lines[-1]
        assert 'reset by peer' in lines[-1]

    def test_connection_closed_after_banner(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'SSH-2.0-ssh-audit-test\r\n')
        vsocket.rdata.append(socket.error(54, 'Connection reset by peer'))
        lines = self._audit(output_spy)
        assert len(lines) == 2
        assert 'error reading packet' in lines[-1]
        assert 'reset by peer' in lines[-1]

    def test_empty_data_after_banner(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'SSH-2.0-ssh-audit-test\r\n')
        lines = self._audit(output_spy)
        assert len(lines) == 2
        assert 'error reading packet' in lines[-1]
        assert 'empty' in lines[-1]

    def test_wrong_data_after_banner(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'SSH-2.0-ssh-audit-test\r\n')
        vsocket.rdata.append(b'xxx\n')
        lines = self._audit(output_spy)
        assert len(lines) == 2
        assert 'error reading packet' in lines[-1]
        assert 'xxx' in lines[-1]

    def test_non_ascii_banner(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'SSH-2.0-ssh-audit-test\xc3\xbc\r\n')
        lines = self._audit(output_spy)
        assert len(lines) == 3
        assert 'error reading packet' in lines[-1]
        assert 'ASCII' in lines[-2]
        assert lines[-3].endswith('SSH-2.0-ssh-audit-test?')

    def test_nonutf8_data_after_banner(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'SSH-2.0-ssh-audit-test\r\n')
        vsocket.rdata.append(b'\x81\xff\n')
        lines = self._audit(output_spy)
        assert len(lines) == 2
        assert 'error reading packet' in lines[-1]
        assert '\\x81\\xff' in lines[-1]

    def test_protocol_mismatch_by_conf(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        vsocket.rdata.append(b'SSH-1.3-ssh-audit-test\r\n')
        vsocket.rdata.append(b'Protocol major versions differ.\n')
        conf = self._conf()
        conf.ssh1, conf.ssh2 = True, False
        lines = self._audit(output_spy, conf)
        assert len(lines) == 4
        assert 'error reading packet' in lines[-1]
        assert 'major versions differ' in lines[-1]



import os
import struct
import pytest

from ssh_audit.auditconf import AuditConf
from ssh_audit.outputbuffer import OutputBuffer
from ssh_audit.protocol import Protocol
from ssh_audit.readbuf import ReadBuf
from ssh_audit.ssh2_kex import SSH2_Kex
from ssh_audit.ssh2_kexparty import SSH2_KexParty
from ssh_audit.ssh_audit import audit
from ssh_audit.writebuf import WriteBuf


# pylint: disable=line-too-long,attribute-defined-outside-init
class TestSSH2:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.OutputBuffer = OutputBuffer
        self.protocol = Protocol
        self.ssh2_kex = SSH2_Kex
        self.ssh2_kexparty = SSH2_KexParty
        self.rbuf = ReadBuf
        self.wbuf = WriteBuf
        self.audit = audit
        self.AuditConf = AuditConf

    def _conf(self):
        conf = self.AuditConf('localhost', 22)
        conf.colors = False
        conf.batch = True
        conf.verbose = True
        conf.ssh1 = False
        conf.ssh2 = True
        return conf

    @classmethod
    def _create_ssh2_packet(cls, payload):
        padding = -(len(payload) + 5) % 8
        if padding < 4:
            padding += 8
        plen = len(payload) + padding + 1
        pad_bytes = b'\x00' * padding
        data = struct.pack('>Ib', plen, padding) + payload + pad_bytes
        return data

    def _kex_payload(self):
        w = self.wbuf()
        w.write(b'\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff')
        w.write_list(['bogus_kex1', 'bogus_kex2'])  # We use a bogus kex, otherwise the host key tests will kick off and fail.
        w.write_list(['ssh-rsa', 'rsa-sha2-512', 'rsa-sha2-256', 'ssh-ed25519'])
        w.write_list(['chacha20-poly1305@openssh.com', 'aes128-ctr', 'aes192-ctr', 'aes256-ctr', 'aes128-gcm@openssh.com', 'aes256-gcm@openssh.com', 'aes128-cbc', 'aes192-cbc', 'aes256-cbc'])
        w.write_list(['chacha20-poly1305@openssh.com', 'aes128-ctr', 'aes192-ctr', 'aes256-ctr', 'aes128-gcm@openssh.com', 'aes256-gcm@openssh.com', 'aes128-cbc', 'aes192-cbc', 'aes256-cbc'])
        w.write_list(['umac-64-etm@openssh.com', 'umac-128-etm@openssh.com', 'hmac-sha2-256-etm@openssh.com', 'hmac-sha2-512-etm@openssh.com', 'hmac-sha1-etm@openssh.com', 'umac-64@openssh.com', 'umac-128@openssh.com', 'hmac-sha2-256', 'hmac-sha2-512', 'hmac-sha1'])
        w.write_list(['umac-64-etm@openssh.com', 'umac-128-etm@openssh.com', 'hmac-sha2-256-etm@openssh.com', 'hmac-sha2-512-etm@openssh.com', 'hmac-sha1-etm@openssh.com', 'umac-64@openssh.com', 'umac-128@openssh.com', 'hmac-sha2-256', 'hmac-sha2-512', 'hmac-sha1'])
        w.write_list(['none', 'zlib@openssh.com'])
        w.write_list(['none', 'zlib@openssh.com'])
        w.write_list([''])
        w.write_list([''])
        w.write_byte(False)
        w.write_int(0)
        return w.write_flush()

    def _kex_payload_with_gss(self):
        w = self.wbuf()
        w.write(b'\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff')
        w.write_list(['gss-gex-sha1-dZuIebMjgUqaxvbF7hDbAw==', 'gss-gex-sha1-vz8J1E9PzLr8b1K+0remTg==', 'gss-group14-sha1-dZuIebMjgUqaxvbF7hDbAw==', 'gss-group14-sha1-vz8J1E9PzLr8b1K+0remTg==', 'gss-group14-sha256-dZuIebMjgUqaxvbF7hDbAw==', 'gss-group14-sha256-vz8J1E9PzLr8b1K+0remTg==', 'gss-group16-sha512-dZuIebMjgUqaxvbF7hDbAw==', 'gss-group16-sha512-vz8J1E9PzLr8b1K+0remTg==', 'gss-group18-sha512-dZuIebMjgUqaxvbF7hDbAw==', 'gss-group18-sha512-vz8J1E9PzLr8b1K+0remTg==', 'gss-group1-sha1-dZuIebMjgUqaxvbF7hDbAw==', 'gss-group1-sha1-vz8J1E9PzLr8b1K+0remTg==', 'gss-curve448-sha512-XXX'])
        w.write_list(['ssh-ed25519'])
        w.write_list(['chacha20-poly1305@openssh.com'])
        w.write_list(['chacha20-poly1305@openssh.com'])
        w.write_list(['hmac-sha2-512-etm@openssh.com'])
        w.write_list(['hmac-sha2-512-etm@openssh.com'])
        w.write_list(['none', 'zlib@openssh.com'])
        w.write_list(['none', 'zlib@openssh.com'])
        w.write_list([''])
        w.write_list([''])
        w.write_byte(False)
        w.write_int(0)
        return w.write_flush()

    def test_kex_read(self):
        kex = self.ssh2_kex.parse(self.OutputBuffer, self._kex_payload())
        assert kex is not None
        assert kex.cookie == b'\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff'
        assert kex.kex_algorithms == ['bogus_kex1', 'bogus_kex2']
        assert kex.key_algorithms == ['ssh-rsa', 'rsa-sha2-512', 'rsa-sha2-256', 'ssh-ed25519']
        assert kex.client is not None
        assert kex.server is not None
        assert kex.client.encryption == ['chacha20-poly1305@openssh.com', 'aes128-ctr', 'aes192-ctr', 'aes256-ctr', 'aes128-gcm@openssh.com', 'aes256-gcm@openssh.com', 'aes128-cbc', 'aes192-cbc', 'aes256-cbc']
        assert kex.server.encryption == ['chacha20-poly1305@openssh.com', 'aes128-ctr', 'aes192-ctr', 'aes256-ctr', 'aes128-gcm@openssh.com', 'aes256-gcm@openssh.com', 'aes128-cbc', 'aes192-cbc', 'aes256-cbc']
        assert kex.client.mac == ['umac-64-etm@openssh.com', 'umac-128-etm@openssh.com', 'hmac-sha2-256-etm@openssh.com', 'hmac-sha2-512-etm@openssh.com', 'hmac-sha1-etm@openssh.com', 'umac-64@openssh.com', 'umac-128@openssh.com', 'hmac-sha2-256', 'hmac-sha2-512', 'hmac-sha1']
        assert kex.server.mac == ['umac-64-etm@openssh.com', 'umac-128-etm@openssh.com', 'hmac-sha2-256-etm@openssh.com', 'hmac-sha2-512-etm@openssh.com', 'hmac-sha1-etm@openssh.com', 'umac-64@openssh.com', 'umac-128@openssh.com', 'hmac-sha2-256', 'hmac-sha2-512', 'hmac-sha1']
        assert kex.client.compression == ['none', 'zlib@openssh.com']
        assert kex.server.compression == ['none', 'zlib@openssh.com']
        assert kex.client.languages == ['']
        assert kex.server.languages == ['']
        assert kex.follows is False
        assert kex.unused == 0

    def _get_empty_kex(self, cookie=None):
        kex_algs, key_algs = [], []
        enc, mac, compression, languages = [], [], ['none'], []
        cli = self.ssh2_kexparty(enc, mac, compression, languages)
        enc, mac, compression, languages = [], [], ['none'], []
        srv = self.ssh2_kexparty(enc, mac, compression, languages)
        if cookie is None:
            cookie = os.urandom(16)
        kex = self.ssh2_kex(self.OutputBuffer, cookie, kex_algs, key_algs, cli, srv, 0)
        return kex

    def _get_kex_variat1(self):
        cookie = b'\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff'
        kex = self._get_empty_kex(cookie)
        kex.kex_algorithms.append('bogus_kex1')
        kex.kex_algorithms.append('bogus_kex2')
        kex.key_algorithms.append('ssh-rsa')
        kex.key_algorithms.append('rsa-sha2-512')
        kex.key_algorithms.append('rsa-sha2-256')
        kex.key_algorithms.append('ssh-ed25519')
        kex.server.encryption.append('chacha20-poly1305@openssh.com')
        kex.server.encryption.append('aes128-ctr')
        kex.server.encryption.append('aes192-ctr')
        kex.server.encryption.append('aes256-ctr')
        kex.server.encryption.append('aes128-gcm@openssh.com')
        kex.server.encryption.append('aes256-gcm@openssh.com')
        kex.server.encryption.append('aes128-cbc')
        kex.server.encryption.append('aes192-cbc')
        kex.server.encryption.append('aes256-cbc')
        kex.server.mac.append('umac-64-etm@openssh.com')
        kex.server.mac.append('umac-128-etm@openssh.com')
        kex.server.mac.append('hmac-sha2-256-etm@openssh.com')
        kex.server.mac.append('hmac-sha2-512-etm@openssh.com')
        kex.server.mac.append('hmac-sha1-etm@openssh.com')
        kex.server.mac.append('umac-64@openssh.com')
        kex.server.mac.append('umac-128@openssh.com')
        kex.server.mac.append('hmac-sha2-256')
        kex.server.mac.append('hmac-sha2-512')
        kex.server.mac.append('hmac-sha1')
        kex.server.compression.append('zlib@openssh.com')
        for a in kex.server.encryption:
            kex.client.encryption.append(a)
        for a in kex.server.mac:
            kex.client.mac.append(a)
        for a in kex.server.compression:
            if a == 'none':
                continue
            kex.client.compression.append(a)
        return kex

    def test_key_payload(self):
        kex1 = self._get_kex_variat1()
        kex2 = self.ssh2_kex.parse(self.OutputBuffer, self._kex_payload())
        assert kex1.payload == kex2.payload

    def test_ssh2_server_simple(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        w = self.wbuf()
        w.write_byte(self.protocol.MSG_KEXINIT)
        w.write(self._kex_payload())
        vsocket.rdata.append(b'SSH-2.0-OpenSSH_7.3 ssh-audit-test\r\n')
        vsocket.rdata.append(self._create_ssh2_packet(w.write_flush()))
        output_spy.begin()
        out = self.OutputBuffer()
        self.audit(out, self._conf())
        out.write()
        lines = output_spy.flush()
        assert len(lines) == 83

    def test_ssh2_server_invalid_first_packet(self, output_spy, virtual_socket):
        vsocket = virtual_socket
        w = self.wbuf()
        w.write_byte(self.protocol.MSG_KEXINIT + 1)
        vsocket.rdata.append(b'SSH-2.0-OpenSSH_7.3 ssh-audit-test\r\n')
        vsocket.rdata.append(self._create_ssh2_packet(w.write_flush()))
        output_spy.begin()
        out = self.OutputBuffer()
        ret = self.audit(out, self._conf())
        out.write()
        assert ret != 0
        lines = output_spy.flush()
        assert len(lines) == 9
        assert 'unknown message' in lines[-1]

    def test_ssh2_gss_kex(self, output_spy, virtual_socket):
        '''Ensure that GSS kex algorithms are properly parsed.'''

        vsocket = virtual_socket
        w = self.wbuf()
        w.write_byte(self.protocol.MSG_KEXINIT)
        w.write(self._kex_payload_with_gss())  # Use the kex with GSS algorithms.
        vsocket.rdata.append(b'SSH-2.0-OpenSSH_7.3 ssh-audit-test\r\n')
        vsocket.rdata.append(self._create_ssh2_packet(w.write_flush()))
        output_spy.begin()
        out = self.OutputBuffer()
        self.audit(out, self._conf())
        out.write()
        lines = output_spy.flush()

        # Ensure that none of the lines are reported as "unknown algorithm".
        for line in lines:
            assert line.find('unknown algorithm') == -1



#!/usr/bin/env python3
"""src/ssh_audit/ssh_audit.py wrapper for backwards compatibility"""

import sys
import traceback
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parent / "src"))

from ssh_audit.ssh_audit import main  # noqa: E402
from ssh_audit import exitcodes  # noqa: E402

exit_code = exitcodes.GOOD

try:
    exit_code = main()
except Exception:
    exit_code = exitcodes.UNKNOWN_ERROR
    print(traceback.format_exc())

sys.exit(exit_code)



import io
import sys
import socket
import pytest


@pytest.fixture(scope='module')
def ssh_audit():
    import ssh_audit.ssh_audit
    return ssh_audit.ssh_audit


# pylint: disable=attribute-defined-outside-init
class _OutputSpy(list):
    def begin(self):
        self.__out = io.StringIO()
        self.__old_stdout = sys.stdout
        sys.stdout = self.__out

    def flush(self):
        lines = self.__out.getvalue().splitlines()
        sys.stdout = self.__old_stdout
        self.__out = None
        return lines


@pytest.fixture(scope='module')
def output_spy():
    return _OutputSpy()


class _VirtualGlobalSocket:
    def __init__(self, vsocket):
        self.vsocket = vsocket
        self.addrinfodata = {}

    # pylint: disable=unused-argument
    def create_connection(self, address, timeout=0, source_address=None):
        # pylint: disable=protected-access
        return self.vsocket._connect(address, True)

    # pylint: disable=unused-argument
    def socket(self,
               family=socket.AF_INET,
               socktype=socket.SOCK_STREAM,
               proto=0,
               fileno=None):
        return self.vsocket

    def getaddrinfo(self, host, port, family=0, socktype=0, proto=0, flags=0):
        key = '{}#{}'.format(host, port)
        if key in self.addrinfodata:
            data = self.addrinfodata[key]
            if isinstance(data, Exception):
                raise data
            return data
        if host == 'localhost':
            r = []
            if family in (0, socket.AF_INET):
                r.append((socket.AF_INET, 1, 6, '', ('127.0.0.1', port)))
            if family in (0, socket.AF_INET6):
                r.append((socket.AF_INET6, 1, 6, '', ('::1', port)))
            return r
        return []


class _VirtualSocket:
    def __init__(self):
        self.sock_address = ('127.0.0.1', 0)
        self.peer_address = None
        self._connected = False
        self.timeout = -1.0
        self.rdata = []
        self.sdata = []
        self.errors = {}
        self.gsock = _VirtualGlobalSocket(self)

    def _check_err(self, method):
        method_error = self.errors.get(method)
        if method_error:
            raise method_error

    def connect(self, address):
        return self._connect(address, False)

    def _connect(self, address, ret=True):
        self.peer_address = address
        self._connected = True
        self._check_err('connect')
        return self if ret else None

    def settimeout(self, timeout):
        self.timeout = timeout

    def gettimeout(self):
        return self.timeout

    def getpeername(self):
        if self.peer_address is None or not self._connected:
            raise OSError(57, 'Socket is not connected')
        return self.peer_address

    def getsockname(self):
        return self.sock_address

    def bind(self, address):
        self.sock_address = address

    def listen(self, backlog):
        pass

    def accept(self):
        # pylint: disable=protected-access
        conn = _VirtualSocket()
        conn.sock_address = self.sock_address
        conn.peer_address = ('127.0.0.1', 0)
        conn._connected = True
        return conn, conn.peer_address

    def recv(self, bufsize, flags=0):
        # pylint: disable=unused-argument
        if not self._connected:
            raise OSError(54, 'Connection reset by peer')
        if not len(self.rdata) > 0:
            return b''
        data = self.rdata.pop(0)
        if isinstance(data, Exception):
            raise data
        return data

    def send(self, data):
        if self.peer_address is None or not self._connected:
            raise OSError(32, 'Broken pipe')
        self._check_err('send')
        self.sdata.append(data)


@pytest.fixture()
def virtual_socket(monkeypatch):
    vsocket = _VirtualSocket()
    gsock = vsocket.gsock
    monkeypatch.setattr(socket, 'create_connection', gsock.create_connection)
    monkeypatch.setattr(socket, 'socket', gsock.socket)
    monkeypatch.setattr(socket, 'getaddrinfo', gsock.getaddrinfo)
    return vsocket



import re
import sys
from setuptools import setup

print_warning = False
m = re.search(r'^VERSION\s*=\s*\'v(\d\.\d\.\d)\'', open('src/ssh_audit/globals.py').read(), re.M)
if m is None:
    # If we failed to parse the stable version, see if this is the development version.
    m = re.search(r'^VERSION\s*=\s*\'v(\d\.\d\.\d-dev)\'', open('src/ssh_audit/globals.py').read(), re.M)
    if m is None:
        print("Error: could not parse VERSION variable from ssh_audit.py.")
        sys.exit(1)
    else:  # Continue with the development version, but print a warning later.
        print_warning = True

version = m.group(1)
print("\n\nPackaging ssh-audit v%s...\n\n" % version)

# see setup.cfg
setup()

if print_warning:
    print("\n\n    !!! WARNING: development version detected (%s).  Are you sure you want to package this version?  Probably not...\n" % version)



import os
import pytest

from ssh_audit.outputbuffer import OutputBuffer
from ssh_audit.ssh2_kex import SSH2_Kex
from ssh_audit.ssh2_kexparty import SSH2_KexParty


@pytest.fixture
def kex(ssh_audit):
    kex_algs, key_algs = [], []
    enc, mac, compression, languages = [], [], ['none'], []
    cli = SSH2_KexParty(enc, mac, compression, languages)
    enc, mac, compression, languages = [], [], ['none'], []
    srv = SSH2_KexParty(enc, mac, compression, languages)
    cookie = os.urandom(16)
    kex = SSH2_Kex(OutputBuffer, cookie, kex_algs, key_algs, cli, srv, 0)
    return kex


def test_prevent_runtime_error_regression(ssh_audit, kex):
    """Prevent a regression of https://github.com/jtesta/ssh-audit/issues/41

    The following test setup does not contain any sensible data.
    It was made up to reproduce a situation when there are several host
    keys, and an error occurred when iterating and modifying them at the
    same time.
    """
    kex.set_host_key("ssh-rsa", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa1", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa2", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa3", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa4", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa5", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa6", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa7", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)
    kex.set_host_key("ssh-rsa8", b"\x00\x00\x00\x07ssh-rsa\x00\x00\x00", 1024, '', 0)

    rv = ssh_audit.build_struct('localhost', None, [], kex=kex)

    assert len(rv["fingerprints"]) == (9 * 2)  # Each host key generates two hash fingerprints: one using SHA256, and one using MD5.

    for key in ['banner', 'compression', 'enc', 'fingerprints', 'kex', 'key', 'mac']:
        assert key in rv



import pytest

from ssh_audit.banner import Banner
from ssh_audit.software import Software


# pylint: disable=attribute-defined-outside-init
class TestVersionCompare:
    @pytest.fixture(autouse=True)
    def init(self, ssh_audit):
        self.software = Software
        self.banner = Banner

    def get_dropbear_software(self, v):
        b = self.banner.parse('SSH-2.0-dropbear_{}'.format(v))
        return self.software.parse(b)

    def get_openssh_software(self, v):
        b = self.banner.parse('SSH-2.0-OpenSSH_{}'.format(v))
        return self.software.parse(b)

    def get_libssh_software(self, v):
        b = self.banner.parse('SSH-2.0-libssh-{}'.format(v))
        return self.software.parse(b)

    def test_dropbear_compare_version_pre_years(self):
        s = self.get_dropbear_software('0.44')
        assert s.compare_version(None) == 1
        assert s.compare_version('') == 1
        assert s.compare_version('0.43') > 0
        assert s.compare_version('0.44') == 0
        assert s.compare_version(s) == 0
        assert s.compare_version('0.45') < 0
        assert s.between_versions('0.43', '0.45')
        assert s.between_versions('0.43', '0.43') is False
        assert s.between_versions('0.45', '0.43') is False

    def test_dropbear_compare_version_with_years(self):
        s = self.get_dropbear_software('2015.71')
        assert s.compare_version(None) == 1
        assert s.compare_version('') == 1
        assert s.compare_version('2014.66') > 0
        assert s.compare_version('2015.71') == 0
        assert s.compare_version(s) == 0
        assert s.compare_version('2016.74') < 0
        assert s.between_versions('2014.66', '2016.74')
        assert s.between_versions('2014.66', '2015.69') is False
        assert s.between_versions('2016.74', '2014.66') is False

    def test_dropbear_compare_version_mixed(self):
        s = self.get_dropbear_software('0.53.1')
        assert s.compare_version(None) == 1
        assert s.compare_version('') == 1
        assert s.compare_version('0.53') > 0
        assert s.compare_version('0.53.1') == 0
        assert s.compare_version(s) == 0
        assert s.compare_version('2011.54') < 0
        assert s.between_versions('0.53', '2011.54')
        assert s.between_versions('0.53', '0.53') is False
        assert s.between_versions('2011.54', '0.53') is False

    def test_dropbear_compare_version_patchlevel(self):
        s1 = self.get_dropbear_software('0.44')
        s2 = self.get_dropbear_software('0.44test3')
        assert s1.compare_version(None) == 1
        assert s1.compare_version('') == 1
        assert s1.compare_version('0.44') == 0
        assert s1.compare_version(s1) == 0
        assert s1.compare_version('0.43') > 0
        assert s1.compare_version('0.44test4') > 0
        assert s1.between_versions('0.44test4', '0.45')
        assert s1.between_versions('0.43', '0.44test4') is False
        assert s1.between_versions('0.45', '0.44test4') is False
        assert s2.compare_version(None) == 1
        assert s2.compare_version('') == 1
        assert s2.compare_version('0.44test3') == 0
        assert s2.compare_version(s2) == 0
        assert s2.compare_version('0.44') < 0
        assert s2.compare_version('0.44test4') < 0
        assert s2.between_versions('0.43', '0.44')
        assert s2.between_versions('0.43', '0.44test2') is False
        assert s2.between_versions('0.44', '0.43') is False
        assert s1.compare_version(s2) > 0
        assert s2.compare_version(s1) < 0

    def test_dropbear_compare_version_sequential(self):
        versions = []
        for i in range(28, 44):
            versions.append('0.{}'.format(i))
        for i in range(1, 5):
            versions.append('0.44test{}'.format(i))
        for i in range(44, 49):
            versions.append('0.{}'.format(i))
        versions.append('0.48.1')
        for i in range(49, 54):
            versions.append('0.{}'.format(i))
        versions.append('0.53.1')
        for v in ['2011.54', '2012.55']:
            versions.append(v)
        for i in range(56, 61):
            versions.append('2013.{}'.format(i))
        for v in ['2013.61test', '2013.62']:
            versions.append(v)
        for i in range(63, 67):
            versions.append('2014.{}'.format(i))
        for i in range(67, 72):
            versions.append('2015.{}'.format(i))
        for i in range(72, 75):
            versions.append('2016.{}'.format(i))
        length = len(versions)
        for i in range(length):
            v = versions[i]
            s = self.get_dropbear_software(v)
            assert s.compare_version(v) == 0
            if i - 1 >= 0:
                vbefore = versions[i - 1]
                assert s.compare_version(vbefore) > 0
            if i + 1 < length:
                vnext = versions[i + 1]
                assert s.compare_version(vnext) < 0

    def test_openssh_compare_version_simple(self):
        s = self.get_openssh_software('3.7.1')
        assert s.compare_version(None) == 1
        assert s.compare_version('') == 1
        assert s.compare_version('3.7') > 0
        assert s.compare_version('3.7.1') == 0
        assert s.compare_version(s) == 0
        assert s.compare_version('3.8') < 0
        assert s.between_versions('3.7', '3.8')
        assert s.between_versions('3.6', '3.7') is False
        assert s.between_versions('3.8', '3.7') is False

    def test_openssh_compare_version_patchlevel(self):
        s1 = self.get_openssh_software('2.1.1')
        s2 = self.get_openssh_software('2.1.1p2')
        assert s1.compare_version(s1) == 0
        assert s2.compare_version(s2) == 0
        assert s1.compare_version('2.1.1') == 0
        assert s2.compare_version('2.1.1p2') == 0
        assert s1.compare_version('2.1.1p1') == 0  # OpenBSD version and p1 version should be the same.
        assert s1.compare_version('2.1.1p2') != 0
        assert s2.compare_version('2.1.1') != 0
        assert s2.compare_version('2.1.1p1') > 0
        assert s2.compare_version('2.1.1p3') < 0
        assert s1.compare_version(s2) != 0
        assert s2.compare_version(s1) != 0

    def test_openbsd_compare_version_sequential(self):
        versions = []
        for v in ['1.2.3', '2.1.0', '2.1.1', '2.2.0', '2.3.0']:
            versions.append(v)
        for v in ['2.5.0', '2.5.1', '2.5.2', '2.9', '2.9.9']:
            versions.append(v)
        for v in ['3.0', '3.0.1', '3.0.2', '3.1', '3.2.2', '3.2.3']:
            versions.append(v)
        for i in range(3, 7):
            versions.append('3.{}'.format(i))
        for v in ['3.6.1', '3.7.0', '3.7.1']:
            versions.append(v)
        for i in range(8, 10):
            versions.append('3.{}'.format(i))
        for i in range(0, 10):
            versions.append('4.{}'.format(i))
        for i in range(0, 10):
            versions.append('5.{}'.format(i))
        for i in range(0, 10):
            versions.append('6.{}'.format(i))
        for i in range(0, 4):
            versions.append('7.{}'.format(i))
        length = len(versions)
        for i in range(length):
            v = versions[i]
            s = self.get_openssh_software(v)
            assert s.compare_version(v) == 0
            if i - 1 >= 0:
                vbefore = versions[i - 1]
                assert s.compare_version(vbefore) > 0
            if i + 1 < length:
                vnext = versions[i + 1]
                assert s.compare_version(vnext) < 0

    def test_libssh_compare_version_simple(self):
        s = self.get_libssh_software('0.3')
        assert s.compare_version(None) == 1
        assert s.compare_version('') == 1
        assert s.compare_version('0.2') > 0
        assert s.compare_version('0.3') == 0
        assert s.compare_version(s) == 0
        assert s.compare_version('0.3.1') < 0
        assert s.between_versions('0.2', '0.3.1')
        assert s.between_versions('0.1', '0.2') is False
        assert s.between_versions('0.3.1', '0.2') is False

    def test_libssh_compare_version_sequential(self):
        versions = []
        for v in ['0.2', '0.3']:
            versions.append(v)
        for i in range(1, 5):
            versions.append('0.3.{}'.format(i))
        for i in range(0, 9):
            versions.append('0.4.{}'.format(i))
        for i in range(0, 6):
            versions.append('0.5.{}'.format(i))
        for i in range(0, 6):
            versions.append('0.6.{}'.format(i))
        for i in range(0, 5):
            versions.append('0.7.{}'.format(i))
        length = len(versions)
        for i in range(length):
            v = versions[i]
            s = self.get_libssh_software(v)
            assert s.compare_version(v) == 0
            if i - 1 >= 0:
                vbefore = versions[i - 1]
                assert s.compare_version(vbefore) > 0
            if i + 1 < length:
                vnext = versions[i + 1]
                assert s.compare_version(vnext) < 0
